from enum import Enum
import torch
from Cache.utils import create_memory
from message_worker import _get_batch_data, _sample_node_neighbors_server, _sample_node_neighbors_single, get_cache_node, pre_sample
from part.Utils import GraphData
from typing import Optional
import torch.distributed as dist
import DistCustomPool
from shared_graph_memory import GraphInfoInShm
import distparser as parser
from torch_geometric.data import Data
import os.path as osp

def partition_load(root: str, algo: str = "metis") -> Data:
    rank = parser._get_worker_rank()
    world_size = parser._get_world_size()
    fn = osp.join(root, f"{algo}_{world_size}", f"{rank:03d}")
    return torch.load(fn)
class DistGraphData(GraphData):
    
    def __init__(self,pdata = None,edge_index = None,path = None,full_edge = True):
        if path is not None: 
            self.rank = parser._get_worker_rank()
        
            path = path + '/rank_' + str(self.rank)
            print('load graph ',path)
            super(DistGraphData,self).__init__(path)
        else:
            #dst和edge在一个分区，src不一定
            #本地节点id
            self.ids = pdata.ids
            self.origin_edge_index = edge_index
            #特征信息
            self.data = Data()
            self.data.x = pdata.x
            self.data.y = pdata.y
            self.data.y = self.data.y.reshape(-1)
            self.data.train_mask = pdata.train_mask
            self.data.test_mask = pdata.test_mask
            self.data.val_mask = pdata.val_mask
            self.rank = parser._get_worker_rank()
            #通信后获得索引
            world_size = parser._get_world_size()
            sample_group = DistCustomPool.get_sample_group()
            self.partition_id = self.rank
            self.partptr = torch.zeros(world_size + 1).int()
            self.num_edge_part = torch.zeros(world_size+1).int()
            self.num_nodes = 0
            self.partitions = world_size
            self.num_parts = self.partitions
            self.num_feasures =  pdata.x.size()[1]
            global_edge = []
            if world_size != 1 and full_edge is not False:
                for rank in range(world_size):
                    #dst在同一个分区，src不一定，映射dst和src的
                    rev_msg = [len(self.ids),len(edge_index[0,:])]
                    print(rev_msg)
                    dist.broadcast_object_list(rev_msg,rank,group=sample_group)
                    self.num_edge_part[rank + 1] = self.num_edge_part[rank] + rev_msg[1]
                    self.num_nodes = self.num_nodes + rev_msg[0]
                    self.partptr[rank + 1] = self.num_nodes
                ptr_index = torch.zeros_like(edge_index)
                self.edgeindex2ptr(self.ids,edge_index[1,:],ptr_index[1,:],self.rank)
                dist.barrier()
                
                for rank in range(world_size):
                    #edge_index映射成新的id
                    if rank != self.rank:
                        rev_idx = torch.zeros(self.partptr[rank+1] - self.partptr[rank]).type_as(self.ids)
                    else:
                        rev_idx = self.ids.clone()
                    print('idx',rev_idx)
                    dist.broadcast(rev_idx,rank,group=sample_group)
                    self.edgeindex2ptr(rev_idx,edge_index[0,:],ptr_index[0,:],rank)
                print(ptr_index)
                self.data.edge_index = ptr_index
                dist.barrier()
                for rank in range(world_size):
                    #edge_index映射成新的id
                    if rank != self.rank:
                        rev_idx = torch.zeros(2,self.num_edge_part[rank + 1] - self.num_edge_part[rank]).type_as(ptr_index)
                    else:
                        rev_idx = ptr_index.clone()
                    print('edge',rev_idx)
                    dist.broadcast(rev_idx,rank,group=sample_group)
                    global_edge.append(rev_idx)
                self.edge_index = torch.cat(global_edge,1)
                dist.barrier()
            else:
                self.num_nodes = len(self.ids)
                self.partptr[1] = self.num_nodes
                self.edge_index = edge_index
                self.data.edge_index = edge_index
            self.num_edges = edge_index[0].numel()

    
    def edgeindex2ptr(self,ids,edge_index,ptr_index,rank):
        #构建映射
        dic = dict(zip(ids.tolist(),torch.arange(len(ids)).tolist()))
        print(dic)
        #values = torch.arange(len(ids))
        #indices = ids
        #sparse_coo = torch.sparse_coo_tensor(indices.unsqueeze(0),values)
        #sparse_csr = sparse_coo.to_sparse_csr
        for i in range(len(edge_index)):
            id = int(edge_index[i].item())
            #print(id)
            if id in dic:
                print(id)
                ptr_index[i] = dic[id] + self.partptr[rank]
        print(ptr_index)                

class DistributedDataLoader:
    ''' 
     Args:
            data_path: the path of loaded graph ,each part 0 of graph is saved on $path$/rank_0
            num_replicas: the num of worker
            

    '''
    def __init__(
            self,
            graph_name,
            graph,
            data_index_mask = None,
            sampler = None,
            collate_fn = None,
            batch_size: Optional[int]=None,
            shuffle:bool = True,
            seed:int = 0,
            drop_last = False,
            device  = 'cpu',
            cache_policy = 'presample',
            cache_memory_size = 0,#cache存储字节大小
            cache_remote_weight = 0,
            **kwargs
    ):
        assert sampler is not None
        self.pool = DistCustomPool.get_sampler_pool()
        self.graph_name = graph_name
        self.batch_size = batch_size
        self.num_workers = parser._get_num_sampler()
        self.queue_size = parser._get_queue_size()
        #if(queue_size is None):
        #    queue_size = self.num_workers * 4 if self.num_workers > 0 else 4
        self.num_pending = 0
        # self.num_node_features = graph.num_node_features
        self.collate_fn = collate_fn
        self.current_pos = 0
        self.drop_last = drop_last
        self.recv_idxs = 0
        self.queue = []
        self.shuffle = shuffle
        self.is_closed = False
        self.sampler = sampler
        self.epoch = 0
        self.graph = graph
        if parser._get_world_size() > 1:
            self.graph_inshm = GraphInfoInShm(graph)
        if self.pool is None:
            self.sampler_info = sampler 
        else :
            self.sampler_info = (sampler.num_nodes,sampler.num_layers,sampler.fanout,sampler.workers)#sampler#._get_sample_info()
        #self.graph_inshm._copy_sampler_to_shame(*sampler._get_neighbors_and_deg())

        self.shuffle=shuffle
        self.seed=seed
        self.kwargs=kwargs
        self.rank = parser._get_worker_rank()
        self.data = torch.arange(start = graph.partptr[self.rank],end = graph.partptr[self.rank+1])
        if(data_index_mask is not None):
            self.data = torch.masked_select(self.data,data_index_mask)
        print('the number of input_number is ',self.data.size(0))
        self._get_expected_idx(drop_last,self.data.size(0))
        
        
        self.device =  device
        self.cache_memory_size = cache_memory_size
        iscached  = self.cache_memory_size != 0 and (self.device=='gpu' or parser._get_world_size()>1)
        
        self.iscached = iscached
        if(self.pool is not None and parser._get_world_size() > 1 ):
            self.pool.set_collate_fn(self.collate_fn,self.sampler_info,self.graph_name,self.graph_inshm)
            dist.barrier()
        elif parser._get_world_size() > 1 :
            self.local_pool = DistCustomPool.LocalSampler()
            self.local_pool.set_collate_fn(self.graph_name,self.graph_inshm,iscached)

        if self.cache_memory_size != 0 and (self.device=='gpu' or parser._get_world_size()>1):
            if cache_policy == 'presample':
                self._pre_sampler(cache_remote_weight)
            #elif cache_policy == 'degree':
                #self._degree_cache(self.data)
            
    def _pre_sampler(self,weight):
        for _ in range(10):
            self.__iter__()
            while self.recv_idxs < self.expected_idx:
            #print(self.recv_idxs)
                batch_data = self._next_data()
        #print(self.graph.partptr[self.rank],self.graph.partptr[self.rank+1])
                pre_sample(self.sampler,batch_data,self.graph_name,weight,self.graph.partptr[self.rank],self.graph.partptr[self.rank+1],self.device)
                self.recv_idxs += 1
        cache_node_list = get_cache_node(self.graph_name,self.cache_memory_size)
        create_memory(self.graph_name,self.graph,self.cache_memory_size,cache_node_list,self.device)
                

    def _get_expected_idx(self,drop_last,data_length):
        if parser._get_world_size() > 1:
            world_size = parser._get_world_size()
            sample_group = DistCustomPool.get_sample_group()
            expected_data = data_length
            for rank in range(world_size):
                len = torch.tensor(data_length)
                dist.broadcast(len,rank,group=sample_group)
                if(drop_last is True):
                    expected_data = min(expected_data,int(len.item()))
                else:
                    expected_data = max(expected_data,int(len.item()))
#
            self.expected_idx = expected_data // self.batch_size
#
            if(not self.drop_last and expected_data % self.batch_size != 0):
                self.expected_idx += 1
            print('expected_index ',self.expected_idx)
#
        else:
            self.expected_idx = data_length// self.batch_size
            if(not self.drop_last and data_length % self.batch_size != 0):
                self.expected_idx += 1


    def __next__(self):
        #print(self.sampler,self.num_samples,self.kwargs)
        #self.sampleGraph(self.sampler,self.num_samplers,self.kwargs)
        #return self.sampleGraph(self.sampler,self.num_samples,**self.kwargs)
        if(self.pool is None and parser._get_world_size() == 1):
            if self.recv_idxs < self.expected_idx:
                batch_data = self._sample_next_batch()
                self.recv_idxs += 1
                assert batch_data is not None
                return batch_data
            else :
                raise StopIteration
        elif self.pool is not None:
            num_reqs = min(self.queue_size - self.num_pending,self.expected_idx - self.submitted)
            for _ in range(num_reqs):
                self._submit_request_next()
                self.submitted = self.submitted + 1
            if self.recv_idxs < self.expected_idx:
                result = self._get_data_from_queue()
                self.recv_idxs += 1
                self.num_pending -=1
                return result
            else :
                assert self.num_pending == 0
                raise StopIteration
        else :
            
            num_reqs = min(self.queue_size - self.num_pending,self.expected_idx - self.submitted)
            for _ in range(num_reqs):
                result = self.local_pool.get_result()
                if result is not None:
                    self.recv_idxs += 1
                    self.num_pending -= 1
                    return result
                else:
                    next_data = self._next_data()
                    if next_data is None:
                        continue
                    assert next_data is not None
                    self.local_pool.sample_next(self.graph_name,self.sampler_info,next_data)
                    self.submitted = self.submitted + 1
                    self.num_pending = self.num_pending + 1
                    
            while(self.recv_idxs < self.expected_idx):
                result = self.local_pool.get_result()
                if result is not None:
                    self.recv_idxs += 1
                    self.num_pending -= 1
                    return result
                
            assert self.num_pending == 0
            raise StopIteration
                    
    def _get_data_from_queue(self,timeout=5000):
        ret = self.pool.get_result(self.graph_name,timeout = timeout)
        return ret
    
    def _sample_next_batch(self):
        next_data = self._next_data()
        if next_data is None:
            return None
        batch_data = _sample_node_neighbors_single(self.graph,self.graph_name,self.sampler,next_data)
        return batch_data

    def _submit_request_next(self):
        next_data = self._next_data()
        if next_data is None:
            return
        self.pool.submit_task(self.graph_name,next_data)
        self.num_pending += 1

    def _next_data(self):
        if self.current_pos >= len(self.data):
            return None
        next_idx = None
        if self.current_pos + self.batch_size > len(self.data):
            if self.drop_last:
                return None
            else:
                next_idx = self.input_data[self.current_pos:]
                self.current_pos = 0
        else:
            next_idx = self.input_data[self.current_pos:self.current_pos + self.batch_size]
            self.current_pos += len(next_idx)
        return next_idx

        
    def __iter__(self):
        if self.shuffle:
            indx = torch.randperm(self.data.size(0))
            self.input_data = self.data[indx]
        else:
            self.input_data = self.data
        #expected_len = self.expected_idx * self.batch_size
        #if(not self.drop_last):
        #    k = expected_len // len(self.input_data)
        #    if(expected_len % len(self.input_data) !=0): k = k+1
        #    if(self.data.shape.numel() == 2):
        #        self.input_data = self.input_data.repeat(k,1)
        #    else:
        #        self.input_data = self.input_data.repeat(k)
        #    self.input_data = self.input_data[:expected_len]
        self.recv_idxs = 0
        self.current_pos = 0
        self.num_pending = 0
        self.submitted = 0
        return self



    def __del__(self):
        if self.pool is not None:
            self.pool.delete_collate_fn(self.graph_name)
    

    def set_epoch(self,epoch):
        self.epoch = epoch 