from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from multiprocessing import Barrier
from multiprocessing.connection import Client
import queue
import sys
import argparse
import traceback
from torch.distributed.rpc import RRef, rpc_async, remote
import torch.distributed.rpc as rpc
import torch
from Cache.cache_memory import get_cpu_cache_data
from part.Utils import GraphData
from Sample.neighbor_sampler import NeighborSampler
import time
from typing import Optional
import torch.distributed as dist
import torch.multiprocessing as map
from BatchData import BatchData
import asyncio 
import os
from Sample.neighbor_sampler import NeighborSampler
from graph_store import _get_graph
from shared_graph_memory import GraphInfoInShm
import distparser as parser
from logger import logger
from countTPS import tps
from Cache.presample_cores import update_count,get_max_rank

WORKER_RANK = parser._get_worker_rank()
NUM_SAMPLER = parser._get_num_sampler()
WORLD_SIZE = parser._get_world_size()
QUEUE_SIZE =parser._get_queue_size()
MAX_QUEUE_SIZE = parser._get_max_queue_size()
RPC_NAME = parser._get_RPC_NAME()



#@rpc.functions.async_execution
def _get_local_attr(data_name,nodes):   
    graph = _get_graph(data_name)
    local_id = graph.get_localId_by_partitionId(parser._get_worker_rank(),nodes)
    return graph.select_attr(local_id)

def _request_remote_attr(rank,data_name,nodes):
    t1  = time.time()
    fut = rpc_async(
        parser._get_RPC_NAME().format(rank),
        _get_local_attr,
        args=(data_name,nodes,)
    )
    #logger.debug('request {}'.format(time.time()-t1))
    return fut
#ThreadPoolExecutor pool 
def _request_all_remote_attr(data_name,nodes_list):
    worker_size = parser._get_world_size()
    worker_rank = parser._get_worker_rank()
    futs = []
    for rank in range(worker_size):
        if(rank == worker_rank):
            futs.append(None)
            continue
        else:
            if(nodes_list[rank].size(0) == 0):
                futs.append(None)
            else:
                futs.append(_request_remote_attr(rank,data_name,nodes_list[rank]))
    return futs

def _check_future_finish(futs):
    check = True
    for _,fut in futs:
        if fut is not None and fut.done() is False:
            check = False
    return check
cnt_remote = 0
total_t0 = 0
total_t1 = 0
total_t2 = 0
total_t3 = 0
total_t4 = 0
cnt = 0
def _split_node_part_and_submit(data_name,node_id_list,iscached):

    graph = _get_graph(data_name)
    worker_size = parser._get_world_size()
    local_rank = parser._get_worker_rank()
    futs = []
    t0 = time.time()
    #local_mask = (graph.partptr[local_rank]<=node_id_list) & (node_id_list<graph.partptr[local_rank+1])
    t1 = time.time()
    local_node = node_id_list[0]#torch.masked_select(node_id_list,local_mask)
    t2 = time.time()
    node_id_list = node_id_list[1]#torch.masked_select(node_id_list,~local_mask)
    t3 = time.time()
    len0 = len(node_id_list)
    #缓存里没有local节点
    if iscached:
        cached_list,node_id_list,value = get_cpu_cache_data(data_name,node_id_list)
    else:
        cached_list = None
        value = None
    t4 = time.time()
    global cnt_remote
    for rank in range(worker_size):
        if(rank != local_rank):
            part_mask = (graph.partptr[rank]<=node_id_list) & (node_id_list<graph.partptr[rank+1])
            part_node = torch.masked_select(node_id_list,part_mask)
            sendlen = 0
            totlen = len(part_node)
            limit_len = 500
            if(part_node.size(0) != 0):
                while(totlen - sendlen >= limit_len):
                    futs.append((part_node[sendlen:sendlen+limit_len],_request_remote_attr(rank,data_name,part_node[sendlen:sendlen+limit_len])))
                    sendlen += limit_len
                if(sendlen < totlen):
                    futs.append((part_node[sendlen:],_request_remote_attr(rank,data_name,part_node[sendlen:])))            
            #if(part_node.size(0) != 0):
            #    futs.append((part_node,_request_remote_attr(rank,data_name,part_node)))
           #tps.count_mess = tps.count_mess + 1
           #tps.t_mask =tps.t_mask + t2 - t1
           #tps.t_rpc = tps.t_rpc + t3 -t2
           #cnt_remote = cnt_remote + len(part_node)
    #tps.print_t2()

    t5 = time.time()
    global cnt
    cnt = cnt + 1
    global total_t0 
    global total_t1 
    global total_t2 
    global total_t3 
    global total_t4 
    total_t0 = total_t0 + t1-t0
    total_t1 = total_t1 + t2-t1
    total_t2 = total_t2 + t3-t2
    total_t3 = total_t3 + t4-t3
    total_t4 = total_t4 + t5-t4
    #logger.debug('t0 {} t1 {} t2 {} t3 {} t4 {}'.format(total_t0/cnt,total_t1/cnt,total_t2/cnt,total_t3/cnt,total_t4/cnt))
    return local_node,futs,cached_list,value

#def _split_node_part(data_name,node_id_list):
#    graph = _get_graph(data_name)
#    part_list=[]
#    scatter_list=[]
#    worker_size = parser._get_world_size()
#    #print(node_id_list)
#    for rank in range(worker_size):
#        group_id=(graph.partptr[rank]<=node_id_list) & (node_id_list<graph.partptr[rank+1])
#        part_list.append(torch.masked_select(node_id_list,group_id))
#        scatter_list.append(torch.nonzero(group_id)) 
#        #print(graph.partptr[rank],graph.partptr[rank+1])
#        #print(torch.nonzero(group_id))
#        #print(node_id_list)
#    scatter_index=torch.cat(scatter_list)
#    # print('part_list',part_list)
#    return part_list,scatter_index
#    
#def _union_remote_and_local(data_name,scatter_list,local_nodes,futs):
#    worker_rank = parser._get_worker_rank()
#    worker_size = parser._get_world_size()
#    feature_part = []
#    local_feature=_get_local_attr(data_name,local_nodes)
#    node_feature_list = []
#    for rank in range(worker_size):
#        if(rank == worker_rank):
#            node_feature_list.append(local_feature)
#        elif(futs[rank] is None):
#            continue
#        else:
#            node_feature_list.append(futs[rank].value())
#
#    node_feature = torch.cat(node_feature_list)
#    attr_size=node_feature.size()
#    x=torch.zeros(attr_size)
#    #print(scatter_list.size())
#    #print(node_feature.size())
#    #print(scatter_list.repeat(1,attr_size[1]).size())
#    #print(x.size())
#    #print(scatter_list)
#    x=x.scatter(0,scatter_list.repeat(1,attr_size[1]),node_feature)
#    return x
#total_local = 0
#total_remote = 0
#def _get_batch_data(kwargs):
#    global total_local
#    global total_remote 
#    data_name = kwargs.get("data_name")
#    input_data = kwargs.get("input_data")
#    nids = kwargs.get("nids")
#    root = nids[:input_data.size(0)]
#    scatter_list,local_nodes,futs = kwargs.get("union_args")
#    edge_index = kwargs.get("edge_index")
#    x=_union_remote_and_local(data_name,scatter_list,local_nodes,futs)
#    graph = _get_graph(data_name)
#    worker_rank = parser._get_worker_rank()
#    local_y_id = graph.get_localId_by_partitionId(worker_rank,root)
#    y = graph.select_y(local_y_id)
#    total_local = total_local + len(local_nodes)
#    total_remote = total_remote + len(x)-len(local_nodes)
#    print('total local',total_local,'total remote',total_remote)
#    #print(t2-t1,t3-t2,t4-t3,t5-t4,time.time()-t4)
#    return BatchData(nids,edge_index,roots=input_data,x=x,y=y)
def _get_batch_data(kwargs):
    data_name = kwargs.get("data_name")
    input_size = kwargs.get("input_size")
    local_nodes,futs,cached,value = kwargs.get("union_args")
    edge_index = kwargs.get("edge_index")
    graph = _get_graph(data_name)
    nids = [local_nodes]
    root = local_nodes[:input_size]
    worker_rank = parser._get_worker_rank()
    local_y_id = graph.get_localId_by_partitionId(worker_rank,root)
    y = graph.select_y(local_y_id)
    x = [_get_local_attr(data_name,local_nodes)]
    if value is not None:
        nids.append(cached)
        x.append(value)
    for (part_node,part_feature) in futs:
        nids.append(part_node)
        x.append(part_feature.value())
    nids = torch.cat(nids,0)
    x = torch.cat(x,0)
    return BatchData(nids,edge_index,roots=root,x=x,y=y)
    
def _sample_node_neighbors_server(data_name,neighbor_nodes,iscached):
    '''
    sample the struct of the subgraph
    '''
    #print('sample node neighbors server')
    t1 = time.time()
    local_node,futs,cached,value = _split_node_part_and_submit(data_name,neighbor_nodes,iscached)
    t2 = time.time()
    #logger.debug('sample server {}'.format(t2-t1))
    return local_node,futs,cached,value

def _sample_node_neighbors_single(graph,graph_name,sampler,input_data):
    nids,edge_index = sampler.sample_from_nodes(input_data.reshape(-1))
    nids = nids[0]
    x= graph.select_attr(nids)
    y = graph.select_y(input_data)
    return BatchData(nids,edge_index,roots=input_data,x=x,y=y)

#存储比例，加载到共享内存,广播缓存
def pre_sample(sampler,input_data,data_name,weight,l,r,device='cpu'):
    #以本地节点为中心进行一次采样
    nodes,_ = sampler.sample_from_nodes(input_data.reshape(-1))
    nodes = torch.cat((nodes[0],nodes[1]),0)
    #print('sample nodes',nodes.masked_select(torch.logical_or(nodes<l,nodes >= r) ))
    update_count(data_name,nodes.reshape(-1),device,weight,l,r)
    
def get_cache_node(data_name,space_size):
    return get_max_rank(data_name,space_size)





